"""
Utilities
"""

import asyncio
from asyncio import Event
from typing import AsyncIterator, Optional, Set, Any

import graphql

from baretypes import (
    Scope,
    HttpRequestCallback,
    HttpMiddlewareCallback
)
from bareasgi.middleware import mw
from graphql import OperationType
from graphql.subscription.map_async_iterator import MapAsyncIterator


async def cancellable_aiter(
        async_iterator: MapAsyncIterator,
        cancellation_event: Event,
        *,
        cancel_pending: bool = True,
        timeout: Optional[float] = None
) -> AsyncIterator:
    """Create a cancellable async iterator.

    :param async_iterator: The async iterator to wrap
    :param cancellation_event: The asyncio Event to controll cancellation.
    :param cancel_pending: If True cancel pending tasks, otherwise wait them.
    :return: The wrapped async iterator
    """
    result_iter = async_iterator.__aiter__()
    cancellation_task = asyncio.create_task(cancellation_event.wait())
    pending: Set[asyncio.Future[Any]] = {
        cancellation_task,
        asyncio.create_task(result_iter.__anext__())
    }

    if timeout is None:
        sleep_task = None
    else:
        sleep_task = asyncio.create_task(asyncio.sleep(timeout))
        pending.add(sleep_task)

    while not cancellation_event.is_set():
        try:
            done, pending = await asyncio.wait(
                pending,
                return_when=asyncio.FIRST_COMPLETED
            )
        except asyncio.CancelledError:
            for pending_task in pending:
                pending_task.cancel()
            raise

        for done_task in done:
            if done_task == cancellation_task:
                for pending_task in pending:
                    if cancel_pending:
                        pending_task.cancel()
                    else:
                        await pending_task
                        yield pending_task.result()
                break
            elif done_task == sleep_task:
                yield None
            else:
                yield done_task.result()
                pending.add(asyncio.create_task(result_iter.__anext__()))
        else:
            if timeout is not None:
                if sleep_task in pending:
                    sleep_task.cancel()
                    pending.discard(sleep_task)
                sleep_task = asyncio.create_task(asyncio.sleep(timeout))
                pending.add(sleep_task)


def _is_http_2(scope: Scope) -> bool:
    return scope['http_version'] in ('2', '2.0')


def _is_subscription(definition: graphql.DefinitionNode) -> bool:
    return isinstance(
        definition,
        graphql.OperationDefinitionNode
    ) and definition.operation is OperationType.SUBSCRIPTION


def has_subscription(document: graphql.DocumentNode) -> bool:
    """Find if a document has a subscription

    :param document: The GraphQL query document
    :type document: graphql.DocumentNode
    :return: True if the document contains a subscription
    :rtype: bool
    """
    return any(_is_subscription(definition) for definition in document.definitions)


def wrap_middleware(
        middleware: Optional[HttpMiddlewareCallback],
        handler: HttpRequestCallback
) -> HttpRequestCallback:
    """Optionally wrap a handler with middleware"""
    if middleware is None:
        return handler
    else:
        return mw(middleware, handler=handler)


class ZeroEvent:
    """An event which blocks when not zero"""

    def __init__(self) -> None:
        self._event = Event()
        self._count = 0
        self._event.set()

    def increment(self) -> int:
        """Increment the count

        :return: The current count
        :rtype: int
        """
        self._count += 1
        self._event.clear()
        return self._count

    def decrement(self) -> int:
        """Decrement the count

        :return: The current count
        :rtype: int
        """
        assert self._count > 0, "Count cannot go below zero"
        self._count -= 1
        if self._count == 0:
            self._event.set()
        return self._count

    async def wait(self) -> None:
        """Wait for the count to be zero"""
        await self._event.wait()

    @property
    def count(self) -> int:
        """The current count

        :return: The current count
        :rtype: int
        """
        return self._count



"""
GraphQL controller
"""

import asyncio
import io
import json
import logging

from cgi import parse_multipart
from datetime import datetime
from typing import List, Dict, Any, Optional, Mapping, cast
from urllib.parse import parse_qs, urlencode

import graphql
import bareutils.header as header

from graphql import GraphQLSchema
from graphql.subscription.map_async_iterator import MapAsyncIterator
from bareasgi import Application
from bareutils import text_reader, text_writer, response_code
from baretypes import (
    Header,
    HttpResponse,
    Scope,
    Info,
    RouteMatches,
    Content,
    WebSocket,
    HttpMiddlewareCallback
)

from .template import make_template
from .websocket_handler import GraphQLWebSocketHandler
from .utils import (
    cancellable_aiter,
    has_subscription,
    wrap_middleware,
    ZeroEvent
)

logger = logging.getLogger(__name__)


def _make_sse_message(val: Optional[graphql.ExecutionResult]) -> str:
    if val is None:
        return f'event: ping\ndata: {datetime.utcnow()}\n\n'

    response = {
        'data': val.data,
        'errors': val.errors
    }

    return f'event: message\ndata: {json.dumps(response)}\n\n'


def _make_json_message(val: Optional[graphql.ExecutionResult]) -> str:
    if val is None:
        return '\n'

    return json.dumps({
        'data': val.data,
        'errors': val.errors
    }) + '\n'


class GraphQLController:
    """GraphQL Controller"""

    def __init__(
            self,
            schema: GraphQLSchema,
            path_prefix: str = '',
            middleware=None,
            ping_interval: float = 10
    ) -> None:
        self.schema = schema
        self.path_prefix = path_prefix
        self.middleware = middleware
        self.ping_interval = ping_interval
        self.ws_subscription_handler = GraphQLWebSocketHandler(schema)
        self.cancellation_event = asyncio.Event()
        self.subscription_count = ZeroEvent()

    def add_routes(
            self,
            app: Application,
            path_prefix: str = '',
            rest_middleware: Optional[HttpMiddlewareCallback] = None,
            view_middleware: Optional[HttpMiddlewareCallback] = None
    ):
        """Add the routes

        :param app: The ASGI application
        :type app: Application
        :param path_prefix: The path prefix
        :type path_prefix: str
        :param rest_middleware: The rest middleware, defaults to None
        :type rest_middleware: Optional[HttpMiddlewareCallback], optional
        :param view_middleware: The view middleware, defaults to None
        :type view_middleware: Optional[HttpMiddlewareCallback], optional
        """
        # Add the REST route
        app.http_router.add(
            {'GET'},
            path_prefix + '/graphql',
            wrap_middleware(rest_middleware, self.handle_graphql)
        )
        app.http_router.add(
            {'POST', 'OPTION'},
            path_prefix + '/graphql',
            wrap_middleware(rest_middleware, self.handle_graphql)
        )
        app.http_router.add(
            {'GET'},
            path_prefix + '/subscriptions',
            wrap_middleware(rest_middleware, self.handle_sse_get)
        )
        app.http_router.add(
            {'POST', 'OPTION'},
            path_prefix + '/subscriptions',
            wrap_middleware(rest_middleware, self.handle_sse_post)
        )

        # Add the subscription route
        app.ws_router.add(
            path_prefix + '/subscriptions',
            self.handle_subscription
        )

        # Add Graphiql
        app.http_router.add(
            {'GET'},
            path_prefix + '/graphiql',
            wrap_middleware(view_middleware, self.view_graphiql)
        )

    async def shutdown(self) -> None:
        """Shutdown the service"""
        self.cancellation_event.set()
        await self.subscription_count.wait()

    # pylint: disable=unused-argument
    async def view_graphiql(
            self,
            scope: Scope,
            info: Info,
            matches: RouteMatches,
            content: Content
    ) -> HttpResponse:
        """Render the Graphiql view"""

        host = header.find(b'host', scope['headers'])
        body = make_template(
            host,
            self.path_prefix + '/graphql',
            self.path_prefix + '/subscriptions'
        )
        headers = [
            (b'content-type', b'text/html'),
            (b'content-length', str(len(body)).encode())
        ]
        return response_code.OK, headers, text_writer(body)

    async def handle_subscription(
            self,
            scope: Scope,
            info: Info,
            matches: RouteMatches,
            web_socket: WebSocket
    ) -> None:
        """Handle a websocket subscription"""
        await self.ws_subscription_handler(scope, info, matches, web_socket)

    @classmethod
    async def _get_query_document(
            cls,
            headers: List[Header],
            content: Content
    ) -> Mapping[str, Any]:
        content_type, parameters = header.content_type(headers)

        if content_type == b'application/graphql':
            return {'query': await text_reader(content)}
        elif content_type in (b'application/json', b'text/plain'):
            return json.loads(await text_reader(content))
        elif content_type == b'application/x-www-form-urlencoded':
            body = parse_qs(await text_reader(content))
            return {name: value[0] for name, value in body.items()}
        elif content_type == b'multipart/form-data':
            return {
                name: value[0]
                for name, value in parse_multipart(
                    io.StringIO(await text_reader(content)),
                    {key.decode('utf-8'): val for key,
                     val in parameters.items()}
                ).items()
            }
        else:
            raise RuntimeError('Content type not supported')

    # pylint: disable=unused-argument
    async def handle_graphql(
            self,
            scope: Scope,
            info: Info,
            matches: RouteMatches,
            content: Content
    ) -> HttpResponse:
        """A request handler for graphql queries"""

        try:
            body = await self._get_query_document(scope['headers'], content)

            query: str = body['query']
            variables: Optional[Dict[str, Any]] = body.get('variables')
            operation_name: Optional[str] = body.get('operationName')

            query_document = graphql.parse(query)

            if has_subscription(query_document):
                method = header.find(b'allow', scope['headers'], b'GET')
                if method == b'GET':
                    # Handle a subscription by returning 201 (Created) with
                    # the url location of the subscription.
                    scheme = scope['scheme']
                    host = header.find(
                        b'host', scope['headers'], b'localhost').decode()
                    path = self.path_prefix + '/subscriptions'
                    query_string = urlencode(
                        {
                            name.encode('utf-8'): json.dumps(value).encode('utf-8')
                            for name, value in body.items()
                        }
                    )
                    location = f'{scheme}://{host}{path}'
                    location += f'?{query_string}'
                    headers = [
                        (b'access-control-expose-headers', b'location'),
                        (b'location', location.encode('ascii'))
                    ]
                    return response_code.CREATED, headers
                else:
                    return await self._handle_sse(scope, info, body)
            else:
                # Handle a query
                result = await graphql.graphql(
                    schema=self.schema,
                    source=graphql.Source(query),  # source=query,
                    variable_values=variables,
                    operation_name=operation_name,
                    context_value=info,
                    middleware=self.middleware
                )

                response: Dict[str, Any] = {'data': result.data}
                if result.errors:
                    response['errors'] = [
                        error.formatted for error in result.errors]

                text = json.dumps(response)
                headers = [
                    (b'content-type', b'application/json'),
                    (b'content-length', str(len(text)).encode())
                ]

                return 200, headers, text_writer(text)

        # pylint: disable=bare-except
        except:
            text = 'Internal server error'
            headers = [
                (b'content-type', b'text/plain'),
                (b'content-length', str(len(text)).encode())
            ]
            return response_code.INTERNAL_SERVER_ERROR, headers, text_writer(text)

    async def handle_sse_get(
            self,
            scope: Scope,
            info: Info,
            matches: RouteMatches,
            content: Content
    ) -> HttpResponse:
        """Handle a server sent event style direct subscription"""

        logger.debug(
            'SSE received GET subscription request: http_version=%s',
            scope['http_version']
        )

        body = {
            name.decode('utf-8'): json.loads(value[0].decode('utf-8'))
            for name, value in parse_qs(scope['query_string']).items()
        }

        return await self._handle_sse(scope, info, body)

    async def handle_sse_post(
            self,
            scope: Scope,
            info: Info,
            matches: RouteMatches,
            content: Content
    ) -> HttpResponse:
        """Handle a server sent event style direct subscription"""

        logger.debug(
            'SSE received POST subscription request: http_version=%s',
            scope['http_version']
        )

        content = await text_reader(content)
        body = json.loads(content)

        return await self._handle_sse(scope, info, body)

    async def _handle_sse(
            self,
            scope: Scope,
            info: Info,
            body: Mapping[str, Any]
    ) -> HttpResponse:
        """Handle a server sent event style direct subscription"""

        accept = header.find(b'accept', scope['headers'], b'text/event-stream')
        content_type = b'application/stream+json' if accept == b'application/json' else accept

        result = cast(
            MapAsyncIterator,
            await graphql.subscribe(
                schema=self.schema,
                document=graphql.parse(body['query']),
                variable_values=body.get('variables'),
                operation_name=body.get('operationName'),
                context_value=info
            )
        )

        # Make an async iterator for the subscription results.

        async def send_events(zero_event: ZeroEvent):
            logger.debug('Started SSE subscription')

            try:
                zero_event.increment()

                async for val in cancellable_aiter(
                        result,
                        self.cancellation_event,
                        timeout=self.ping_interval
                ):
                    print(val)
                    if content_type == b'text/event-stream':
                        yield _make_sse_message(val).encode('utf-8')
                        # Give the ASGI server a nudge.
                        yield b':\n\n'
                    else:
                        yield _make_json_message(val).encode('utf-8')
                        yield b'\n'

            except asyncio.CancelledError:
                logger.debug("Cancelled SSE subscription")
            finally:
                zero_event.decrement()

            logger.debug('Stopped SSE subscription')

        headers = [
            (b'cache-control', b'no-cache'),
            # (b'content-type', b'text/event-stream'),
            (b'content-type', content_type),
            (b'connection', b'keep-alive')
        ]

        return response_code.OK, headers, send_events(self.subscription_count)



"""
GraphQL controller
"""

import logging

from typing import Optional

from graphql import GraphQLSchema
from bareasgi import Application
from baretypes import (
    Scope,
    Info,
    HttpMiddlewareCallback
)

from .controller import GraphQLController

logger = logging.getLogger(__name__)


def add_graphql_next(
        app: Application,
        schema: GraphQLSchema,
        path_prefix: str = '',
        rest_middleware: Optional[HttpMiddlewareCallback] = None,
        view_middleware: Optional[HttpMiddlewareCallback] = None,
        graphql_middleware=None,
        ping_interval: float = 10
) -> None:
    """Add graphql support to an bareASGI application.

    :param app: The bareASGI application.
    :param schema: The GraphQL schema to use.
    :param path_prefix: An optional path prefix from which to provide endpoints.
    :param rest_middleware: Middleware for the rest end points.
    :param view_middleware: Middleware from the GraphiQL end point.
    :param graphql_middleware: Middleware for graphql-core-next.
    :param subscription_expiry: The time to wait before abandoning an unused subscription.
    :return: Returns the constructed controller.
    """

    # pylint: disable=unused-argument
    async def start_graphql(scope: Scope, info: Info, request) -> None:
        """Start the GraphQL controller"""

        logger.debug('Starting the GraphQL controller')

        controller = GraphQLController(
            schema,
            path_prefix,
            graphql_middleware,
            ping_interval
        )
        controller.add_routes(app, path_prefix, rest_middleware, view_middleware)
        info['graphql_controller'] = controller

    # pylint: disable=unused-argument
    async def stop_graphql(scope: Scope, info: Info, request) -> None:
        """Stop the GraphQL controller"""

        logger.debug('Stopping the GraphQL controller')

        graphql_controller: GraphQLController = info['graphql_controller']
        await graphql_controller.shutdown()

    app.startup_handlers.append(start_graphql)
    app.shutdown_handlers.append(stop_graphql)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['bareasgi_graphql_next']

package_data = \
{'': ['*']}

install_requires = \
['bareASGI>=3.0,<4.0', 'bareutils>=3.0,<4.0', 'graphql-core>=3.0,<4.0']

setup_kwargs = {
    'name': 'bareasgi-graphql-next',
    'version': '3.9.2',
    'description': 'GraphQL support for the bareASGI framework',
    'long_description': '# bareASGI-graphql-next\n\nGraphql support for [bareASGI](http://github.com/rob-blackbourn/bareasgi) (read the [documentation](https://bareasgi-graphql-next.readthedocs.io/en/latest/))\n\nThe controller provides a GraphQL GET and POST route, a WebSocket subscription server, and a Graphiql view.\n\n## Installation\n\nInstall from the pie shop.\n\n```bash\npip install bareasgi-graphql-next\n```\n\n## Usage\n\nYou can register the graphql controller with the `add_graphql_next` function.\n\n```python\nfrom bareasgi import Application\nfrom bareasgi_graphql_next import add_graphql_next\nimport graphql\n\n# Get the schema ...\nschema = graphql.GraphQLSchema( ... )\n\nimport uvicorn\n\napp = Application()\nadd_graphql_next(app, schema)\n\nuvicorn.run(app, port=9009)\n\n```\n\n',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'url': 'https://github.com/rob-blackbourn/bareasgi-graphql-next',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
